
@file:OptIn(ExperimentalMaterial3Api::class)

package com.example.app.screens

import android.app.Application
import android.util.Log
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Delete
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.HorizontalDivider
import androidx.compose.material3.Icon
import androidx.compose.material3.OutlinedIconButton
import androidx.compose.material3.Scaffold
import androidx.compose.material3.SegmentedButton
import androidx.compose.material3.SegmentedButtonDefaults
import androidx.compose.material3.SingleChoiceSegmentedButtonRow
import androidx.compose.material3.Text
import androidx.compose.material3.TopAppBar
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.unit.dp
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.viewModelScope
import androidx.lifecycle.viewmodel.compose.viewModel
import com.arcgismaps.geometry.AreaUnit
import com.arcgismaps.geometry.GeodeticCurveType
import com.arcgismaps.geometry.GeometryEngine
import com.arcgismaps.geometry.GeometryType
import com.arcgismaps.geometry.LinearUnit
import com.arcgismaps.geometry.MeasurementUnit
import com.arcgismaps.geometry.Point
import com.arcgismaps.geometry.Polygon
import com.arcgismaps.geometry.Polyline
import com.arcgismaps.geometry.SpatialReference
import com.arcgismaps.mapping.ArcGISMap
import com.arcgismaps.mapping.BasemapStyle
import com.arcgismaps.mapping.Viewpoint
import com.arcgismaps.mapping.view.geometryeditor.GeometryEditor
import com.arcgismaps.mapping.view.geometryeditor.VertexTool
import com.arcgismaps.toolkit.geoviewcompose.MapView
import com.example.app.R
import kotlinx.coroutines.launch
import kotlin.math.round

class MapViewModel(application: Application) : AndroidViewModel(application) {

    // Create a map using the basemap style ArcGISTopographic.
    val map = ArcGISMap(BasemapStyle.ArcGISTopographic).apply {
        initialViewpoint = Viewpoint(
            center = Point(
                x = -1.880843,
                y = 53.479979,
                spatialReference = SpatialReference.wgs84()
            ),
            scale = 10_000_000.0
        )
    }

    val geometryEditor = GeometryEditor()

    // Create mutable state variables for displaying the current planar and geodetic measurements.
    var geodeticMeasurement by mutableStateOf("0")
    var planarMeasurement by mutableStateOf("0")
    var measureType by mutableStateOf("measurement")
    var displayUnits by mutableStateOf("")

    /**
     * Reset the strings that display measurements.
     */
    fun resetDisplayText() {
        displayUnits = ""
        measureType = "measurement"
        geodeticMeasurement = "0"
        planarMeasurement = "0"
    }

    /**
     * Log errors.
     */
    private fun logError(error: Throwable) {
        Log.e(this.javaClass.simpleName, error.message.toString(), error.cause)
    }

    /**
     * Starts the GeometryEditor using the selected [GeometryType].
     */
    fun startEditor(selectedGeometry: GeometryType) {
        geometryEditor.apply {
            // Stops the current editing session
            stop()
            // Start new editing session
            start(selectedGeometry)
        }
    }

    /**
     * Deletes the selected element and stops the geometry editor if there are no
     * more elements in the geometry.
     */
    fun deleteSelection() {
        if (geometryEditor.geometry.value?.isEmpty == true) {
            geometryEditor.stop()
            return
        }

        // Select the entire geometry instead of the just the most recent edit.
        geometryEditor.selectGeometry()

        val selectedElement = geometryEditor.selectedElement.value
        if (selectedElement?.canDelete == true) {
            geometryEditor.deleteSelectedElement()
        }

        resetDisplayText()
    }

    /**
     * Calculates the geometry's measurement.
     */
    private fun calculateMeasurements() {

        // Retrieve the latest state of the geometry using the geometry editor.
        val geometryToMeasure = geometryEditor.geometry.value
            ?: return logError(Exception("Geometry passed is null."))

        // Simplify the geometry to make it topologically consistent according to its type.
        // Among other things, simplifying ensures that areas are never returned as negative values.
        val simplifiedGeometry = GeometryEngine.simplifyOrNull(geometry = geometryToMeasure)
            ?: return logError(Exception("Failed to simplify the geometry"))

        val spatialReferenceOfGeometry = simplifiedGeometry.spatialReference
            ?: return logError(Exception("The geometry has no spatial reference"))

        when (simplifiedGeometry) {

            is Polyline -> {
                // Get the geodetic length.
                val lengthGeodetic = GeometryEngine.lengthGeodetic(
                    geometry = simplifiedGeometry,
                    lengthUnit = LinearUnit.kilometers,
                    curveType = GeodeticCurveType.Geodesic
                )
                // Get the planar length, which is returned in meters.
                val lengthPlanar = GeometryEngine.length(geometry = simplifiedGeometry)

                // Update UI fields using the GeometryEngine results.
                displayMeasurements(
                    measurementGeodetic = lengthGeodetic,
                    measurementPlanar = lengthPlanar,
                    geometryType = GeometryType.Polyline,
                    spatialReferenceUnit = spatialReferenceOfGeometry.unit
                )
            }

            is Polygon -> {
                // Get the geodetic area.
                val areaGeodetic = GeometryEngine.areaGeodetic(
                    geometry = simplifiedGeometry,
                    unit = AreaUnit.squareKilometers,
                    curveType = GeodeticCurveType.Geodesic
                )

                // Get the planar area, which is returned in square meters.
                val areaPlanar = GeometryEngine.area(geometry = simplifiedGeometry)

                // Update UI fields using the GeometryEngine results.
                displayMeasurements(
                    measurementGeodetic = areaGeodetic,
                    measurementPlanar = areaPlanar,
                    geometryType = GeometryType.Polygon,
                    spatialReferenceUnit = spatialReferenceOfGeometry.unit
                )
            }
            else -> {}
        }

    }


    /**
     * Rounds number to 3 decimal places.
     */
    private fun Double.roundToThreeDecimals(): Double {
        return round(this * 1000.0) / 1000.0
    }

    private fun displayMeasurements(
        measurementGeodetic: Double,
        measurementPlanar: Double,
        geometryType: GeometryType,
        spatialReferenceUnit: MeasurementUnit
    ) {

        when (geometryType) {
            GeometryType.Polyline -> {
                displayUnits = "km"
                measureType = "length"

                // Round the geodetic measurement. Then display it.
                geodeticMeasurement = measurementGeodetic.roundToThreeDecimals().toString()

                // If a spatial reference is projected, its measurement unit is LinearUnit. In this tutorial, you can tell
                // the spatial reference by inspection. However, for more complex apps, it is good practice to check.
                if (spatialReferenceUnit is LinearUnit) {
                    val convertedPlanarMeasurement = spatialReferenceUnit.convertTo(
                        toUnit = LinearUnit.kilometers,
                        value = measurementPlanar
                    )
                    planarMeasurement = convertedPlanarMeasurement.roundToThreeDecimals().toString()
                }
            }

            GeometryType.Polygon -> {
                displayUnits = "km²"
                measureType = "area"

                // Round the geodetic measurement. Then display it.
                geodeticMeasurement = measurementGeodetic.roundToThreeDecimals().toString()

                // If a spatial reference is projected, its measurement unit is LinearUnit. In this tutorial, you can tell
                // the spatial reference by inspection. However, for more complex apps, it is good practice to check.
                if (spatialReferenceUnit is LinearUnit) {
                    val convertedPlanarMeasurement = AreaUnit(spatialReferenceUnit).convertTo(
                        toUnit = AreaUnit.squareKilometers,
                        area = measurementPlanar
                    )
                    planarMeasurement = convertedPlanarMeasurement.roundToThreeDecimals().toString()
                }
            }

            else -> {}

        }

    }

    /**
     * Load the map and set GeometryEditor tool after the map is loaded.
     */
    init {
        viewModelScope.launch {
            // Load the map.
            map.load().onFailure { error ->
                logError(error)
            }
            // Set Geometry Editor tool when map is loaded.
            geometryEditor.tool = VertexTool()
            viewModelScope.launch {
                geometryEditor.geometry.collect {
                    // Update the length/area measurement as the geometry changes.
                    calculateMeasurements()
                }
            }
        }
    }

}

@Composable
fun MainScreen() {

    // Create a ViewModel to handle MapView interactions.
    val mapViewModel: MapViewModel = viewModel()

    Scaffold(topBar = { TopAppBar(title = { Text(text = stringResource(id = R.string.app_name)) }) }) {
        Column(
            modifier = Modifier
                .fillMaxSize()
                .padding(it)
        ) {
            MapView(
                modifier = Modifier
                    .fillMaxSize()
                    .weight(1f),
                arcGISMap = mapViewModel.map,
                geometryEditor = mapViewModel.geometryEditor
            )

            GeometryOptions(mapViewModel)
        }
    }

}

@Composable
fun GeometryOptions(mapViewModel: MapViewModel) {

    Column(modifier = Modifier.padding(24.dp), verticalArrangement = Arrangement.spacedBy(12.dp)) {
        Text(text = "Geodesic ${mapViewModel.measureType}: ${mapViewModel.geodeticMeasurement} ${mapViewModel.displayUnits}")
        Text(text = "Planar ${mapViewModel.measureType}: ${mapViewModel.planarMeasurement} ${mapViewModel.displayUnits}")
        HorizontalDivider()

        Row(
            modifier = Modifier.fillMaxWidth(),
            horizontalArrangement = Arrangement.SpaceEvenly
        ) {
            // Default to an unselected state.
            var selectedIndex by remember { mutableIntStateOf(-1) }
            // A two-choice button: Polyline and Polygon tools.
            SingleChoiceSegmentedButtonRow {
                // Create segmented buttons for each Geometry.
                listOf("Polyline", "Polygon").forEachIndexed { index, label ->
                    SegmentedButton(
                        shape = SegmentedButtonDefaults.itemShape(index = index, count = 2),
                        selected = index == selectedIndex,
                        onClick = {
                            // Update the selected index.
                            selectedIndex = index
                            // Reset UI text fields.
                            mapViewModel.resetDisplayText()
                            // Start the GeometryEditor
                            mapViewModel.startEditor(
                                selectedGeometry = if (index == 0) GeometryType.Polyline else GeometryType.Polygon
                            )
                        }
                    ) { Text(label) }
                }
            }
            // The Delete tool uses the default delete icon: trash can.
            OutlinedIconButton(onClick = mapViewModel::deleteSelection) {
                Icon(
                    imageVector = Icons.Default.Delete,
                    contentDescription = "Delete"
                )
            }
        }

    }

}


